(* Complex Vector Arithmetic:
   solving the Shonan challenge by Suda-sensei

   https://github.com/StagedHPC/shonan-challenge/blob/master/problems/

``If a code written for one of the representation [of complex vectors]
and a new data representation are given, can we generate code for the
new data representation?''

*)

(*{{{ Let's be systematic now. Let's develop a DSL *)

(* ------------------------------------------------------------------------
   Part 1: Abstracting over numbers
*)

open Ring
(*
 #load "ring.cmo";;
*)

(* Complex numbers, parameterized over the real number representation *)
open Cmplx
(* type 'a cmplx = {re: 'a; im: 'a} *)

(* To be domain of vectors, cmplx must be a RING *)
(* QQQ: What if we forget sub? *)
module RingCmplx(R:RING) = struct
  type t = R.t cmplx
  let zero = {re = R.zero; im = R.zero}
  let one  = {re = R.one;  im = R.zero}
  let add = fun x y ->
    {re = R.add x.re y.re;
     im = R.add x.im y.im}
  let sub = fun x y ->
    {re = R.sub x.re y.re;
     im = R.sub x.im y.im}
  let mul = fun x y ->
    {re = R.sub (R.mul x.re y.re) (R.mul x.im y.im);
     im = R.add (R.mul x.re y.im) (R.mul x.im y.re)}
end;;

(* Now, float cmplx is a RING too *)

(* Adding abbreviations since FloatComplex will be used often *)
module FloatComplex = RingCmplx(RingFloat)
type float_complex = FloatComplex.t

let {re = 4.; im = 4.} =
  let f = let module M = ExArith(FloatComplex) in M.xsq1
  in f {re=2.0; im=1.0}  (* test *)
;;


module FloatCodeComplex = RingCmplx(RingFloatCode)
type float_code_complex = FloatCodeComplex.t

(* QQQ What is the purpose of those functions? *)
(* QQQ Think of let-insertion on x *)
let of_complex_code : float cmplx code -> float_code_complex =
  fun x -> {re = .<.~x.re>.; im = .<.~x.im>.}
let of_code_complex : float_code_complex -> float cmplx code =
  fun {re;im} -> .<{re = .~re; im = .~im}>.

let _ = 
  let f = let module M = ExArith(FloatCodeComplex) in M.xsq1
  in .<fun x -> 
         .~(of_code_complex 
              (f (of_complex_code .<x>.)))>.
;;
(*
      - : (float Cmplx.cmplx -> float Cmplx.cmplx) code = .<
fun x_3  ->
  {
    Cmplx.im =
      (((x_3.Cmplx.re *. x_3.Cmplx.im) +. (x_3.Cmplx.im *. x_3.Cmplx.re)) +.
         0.);
    Cmplx.re =
      (((x_3.Cmplx.re *. x_3.Cmplx.re) -. (x_3.Cmplx.im *. x_3.Cmplx.im)) +.
         1.)
  }>.

*)

(* QQQ Consider a different staging of complex numbers:
   float cmplx code. What are advantages and disadvantages?
*)

(* ------------------------------------------------------------------------
   Part 2: Abstracting over vectors
 *)

open Vector
(*
 #load "vector.cmo";;
*)


(* ------------------------------------------------------------------------
   Very simple BLAS Level 1: vector assignment and element-wise multiplications
 *)

(* See BLAS1 in vector.ml *)

(* element-wise vector multiplication problem itself *)
module VMULT(R: RING)(A: VEC) = struct
  module L = BLAS1(R)(A)
  open L
  let vmult vout v1 v2 = 
    vout := v1 *. v2
end

(*  That was very general, textbook code, which we write once and for all.
    After that, we instantiate it in various ways, as below.
*)

(* Let us try it out, on ordinary complex arrays *)

let vmult : 
    float_complex array -> float_complex array -> float_complex array -> unit =
  fun vout v1 v2 ->
    let n = Array.length vout in  
                           (* vector representations *)
    let vout = OVec (n, fun i v -> vout.(i) <- v) in
    let v1   = Vec (n,fun i -> v1.(i)) in
    let v2   = Vec (n,fun i -> v2.(i)) in

    let module V = VMULT(FloatComplex)(VecSta) in
    V.vmult vout v1 v2
;;


(*{{{ Tests *)

(* Does it actually work? Let's try out. Here's the unit test *)
let v1 = [| {re=1.;im=0.}; {re=0.;im=1.;}; {re=2.;im=3.}; {re=3.;im=4.}; |];;
let v2 = [| {re=0.;im=1.}; {re=6.;im=7.;}; {re=0.;im=0.}; {re=1.;im=2.}; |];;
let vout = 
  let vout = Array.make (Array.length v1) {re=0.;im=0.} in
  vmult vout v1 v2; vout
;;

let  [|{re = 0.; im = 1.}; {re = -7.; im = 6.}; {re = 0.; im = 0.};
       {re = -5.; im = 10.}|]
 = vout

(*}}}*)

(* The classic case of lots of abstraction and lots of guilt.
   Lots of functors, which involve indirection, lots of array access
   indirection. The performance will be really bad.

   To improve, we `compile', or generate the code.
*)

(*}}}*)

(*}}}*)

(*{{{ Generating code for float cmplx vector multiplication *)

(* ========================================================================
   Code Generation
*)


(* ------------------------------------------------------------------------
   Generating code for the ordinary complex vector multiplication
   We use the same VMULT, and essentially the same vmult, but
   instantiate them differently.
*)

(* QQQ Suggestions? *)


(*{{{ Lifting vectors *)

(* See VecDyn in vector.ml *)


(*}}}*)

(* Code for vmult_ca
   Start with vmult, set the desired signature and then adjust the code
*)
(*{{{ Staged vmult *)

let vmult_ca : 
    (float_complex array -> float_complex array -> float_complex array -> unit)
    code =
 .<fun vout v1 v2 ->
    let n = Array.length vout in  
                                      (* vector representations *)
  .~(let vout = OVec (.<n>., fun i v -> 
       .<vout.(.~i) <- .~(of_code_complex v)>.) in
     let v1   = Vec  (.<n>., fun i -> 
       of_complex_code .<v1.(.~i)>.) in
     let v2   = Vec  (.<n>., fun i -> 
       of_complex_code .<v2.(.~i)>.) in

     (* let module R =  *)
     (*      RingCmplxCode(struct type base = float end)(RingFloatCode) in *)
     let module V = VMULT(FloatCodeComplex)(VecDyn) in
     V.vmult vout v1 v2)
  >.
;;
(*
val vmult_ca :
  (float_complex array -> float_complex array -> float_complex array -> unit)
  code = .<
  fun vout_4  ->
    fun v1_5  ->
      fun v2_6  ->
        let n_7 = Array.length vout_4  in
        for i_8 = 0 to n_7 - 1 do
          vout_4.(i_8) <-
            {
              Cmplx.im =
                (((v1_5.(i_8)).Cmplx.re *. (v2_6.(i_8)).Cmplx.im) +.
                   ((v1_5.(i_8)).Cmplx.im *. (v2_6.(i_8)).Cmplx.re));
              Cmplx.re =
                (((v1_5.(i_8)).Cmplx.re *. (v2_6.(i_8)).Cmplx.re) -.
                   ((v1_5.(i_8)).Cmplx.im *. (v2_6.(i_8)).Cmplx.im))
            }
        done>.
*)
(* Need let-insertion *)

open Runcode

let  [|{re = 0.; im = 1.}; {re = -7.; im = 6.}; {re = 0.; im = 0.};
       {re = -5.; im = 10.}|] =
  let vout = Array.make (Array.length v1) {re=0.;im=0.} in
  run vmult_ca vout v1 v2; vout

(*}}}*)


(* That was representation of complex vectors as arrays of structures.
   What about a different representation, structure of arrays?
*)

(*}}}*)

(* Let's pass real and complex parts as separate arrays *)
(* Again, we start with the signature, and let the code write itself *)
(*{{{ Instead of vector of structures, use structures of vectors *)

let vmult_ac : 
    (float array cmplx -> 
     float array cmplx -> 
     float array cmplx -> unit)
    code =
 .<fun vout v1 v2 ->
    let n = Array.length vout.re in
    let v1re = v1.re and v1im = v1.im and v2re = v2.re and v2im = v2.im in

                                     (* QQQ add assertions about dimensions *)
                                     (* vector representations *)
  .~(let vout = OVec (.<n>., fun i v -> 
                             .<begin vout.re.(.~i) <- .~(v.re);
                                     vout.im.(.~i) <- .~(v.im); end>.) in
     let v1   = Vec  (.<n>., fun i -> 
                              {re= .<v1re.(.~i)>.;im= .<v1im.(.~i)>.}) in
     let v2   = Vec  (.<n>., fun i -> 
                              {re= .<v2re.(.~i)>.;im= .<v2im.(.~i)>.}) in

     let module V = VMULT(FloatCodeComplex)(VecDyn) in
     V.vmult vout v1 v2)
  >.
;;
(*
val vmult_ac :
  (float array Cmplx.cmplx ->
   float array Cmplx.cmplx -> float array Cmplx.cmplx -> unit)
  code = .<
  fun vout_9  ->
    fun v1_10  ->
      fun v2_11  ->
        let n_12 = Array.length vout_9.Cmplx.re  in
        let v1re_13 = v1_10.Cmplx.re
        
        and v1im_14 = v1_10.Cmplx.im
        
        and v2re_15 = v2_11.Cmplx.re
        
        and v2im_16 = v2_11.Cmplx.im
         in
        for i_17 = 0 to n_12 - 1 do
          (vout_9.Cmplx.re).(i_17) <-
            (((v1re_13.(i_17)) *. (v2re_15.(i_17))) -.
               ((v1im_14.(i_17)) *. (v2im_16.(i_17))));
          (vout_9.Cmplx.im).(i_17) <-
            (((v1re_13.(i_17)) *. (v2im_16.(i_17))) +.
               ((v1im_14.(i_17)) *. (v2re_15.(i_17))))
        done>.
  
*)

let {re = [|0.; -7.; 0.; -5.|]; im = [|1.; 6.; 0.; 10.|]}
 =
  let v1c   = {re = Array.init (Array.length v1) (fun i -> v1.(i).re);
              im = Array.init (Array.length v1) (fun i -> v1.(i).im) } in
  let v2c   = {re = Array.init (Array.length v2) (fun i -> v2.(i).re);
              im = Array.init (Array.length v2) (fun i -> v2.(i).im) } in
  let vout = {re = Array.make (Array.length v1) 0.; 
              im = Array.make (Array.length v1) 0.} in
  run vmult_ac vout v1c v2c; vout


(* QQQ How to avoid repeated indexing, as in v1_re_22.(i_27)?
   That is, how to insert let?
*)

(*}}}*)

(* EXERCISE: different representation: the first half of the array
   is the real part, the second part: imaginary
*)

(*{{{ First half: Re, second part: Im *)

let vmult_ac2 : 
    (float array -> float array -> float array -> unit)
    code =
 .<fun vout_ri v1_ri v2_ri ->
    let n2 = Array.length vout_ri in  
    let n  = n2 / 2 in
                                     (* QQQ add assertions about dimensions *)
                                     (* vector representations *)
  .~(let vout = OVec (.<n>., fun i v -> 
                             .<begin vout_ri.(.~i)   <- .~(v.re);
                                     vout_ri.(.~i+n) <- .~(v.im); end>.) in
     let v1   = Vec  (.<n>., fun i -> 
                              {re= .<v1_ri.(.~i)>.;im= .<v1_ri.(.~i+n)>.}) in
     let v2   = Vec  (.<n>., fun i -> 
                              {re= .<v2_ri.(.~i)>.;im= .<v2_ri.(.~i+n)>.}) in

     let module R = RingCmplx(RingFloatCode) in
     let module V = VMULT(R)(VecDyn) in
     V.vmult vout v1 v2)
  >.
;;
(*
val vmult_ac2 : (float array -> float array -> float array -> unit) code = .<
  fun vout_ri_16  v1_ri_17  v2_ri_18  ->
    let n2_19 = Array.length vout_ri_16 in
    let n_20 = n2_19 / 2 in
    for i_21 = 0 to n_20 - 1 do
      vout_ri_16.(i_21) <-
      ((v1_ri_17.(i_21)) *. (v2_ri_18.(i_21))) -.
        ((v1_ri_17.(i_21 + n_20)) *. (v2_ri_18.(i_21 + n_20)));
      vout_ri_16.(i_21 + n_20) <-
      ((v1_ri_17.(i_21)) *. (v2_ri_18.(i_21 + n_20))) +.
        ((v1_ri_17.(i_21 + n_20)) *. (v2_ri_18.(i_21)))
    done>.
*)
let ([|0.; -7.; 0.; -5.; 1.; 6.; 0.; 10.|])
 =
  let n = (Array.length v1) in
  let v1_ri = Array.init (2*n) (fun i -> 
     if i < n then v1.(i).re else v1.(i-n).im) in
  let v2_ri = Array.init (2*n) (fun i -> 
     if i < n then v2.(i).re else v2.(i-n).im) in
  let vout_ri = Array.make (2*n) 0. in
  !. vmult_ac2 vout_ri v1_ri v2_ri; vout_ri

(*}}}*)

(*
 EXERCISE: modify the code for the case of polar representation of
  complex numbers

EXERCISE:
  Generate code for other complex vector operations (for example,
  computing $a \star x_j + y_j$ where $a$ is a scalar)
  The common BLAS name for this operation is caxpy ('p' stands for plus)
*)

(* Try incremental improvement, reusing the old code! *)

(* Extend LA to add scalar-vector multiplication *)
module BLAS1M(R: RING)(A: VEC) = struct
  include BLAS1(R)(A)
  open R
  let ( +. ) v1 v2 = zip_with add v1 v2	     (* element-wise add *)
  let ( *% ) s v = vec_map (fun x -> mul s x) v
end

(* The output of OCaml shows the fully extended structure *)

(* element-wise vector multiplication problem itself *)
module CAXPY(R: RING)(A: VEC) = struct
  module L = BLAS1M(R)(A)
  open L
  let caxpy vout s v1 v2 = 
    vout := s *% v1 +. v2
end

let caxpy_ca : 
    (float_complex array -> float_complex -> float_complex array -> 
      float_complex array -> unit)
    code =
 .<fun vout s v1 v2 ->
    let n = Array.length vout in  
                                      (* vector representations *)
  .~(let vout = OVec (.<n>., fun i v -> 
       .<vout.(.~i) <- .~(of_code_complex v)>.) in
     let v1   = Vec  (.<n>., fun i -> 
       of_complex_code .<v1.(.~i)>.) in
     let v2   = Vec  (.<n>., fun i -> 
       of_complex_code .<v2.(.~i)>.) in

     let module V = CAXPY(FloatCodeComplex)(VecDyn) in
     V.caxpy vout (of_complex_code .<s>.) v1 v2)
  >.
;;

let _ = print_endline "All Done";;
