(* The staged interpreter for the simple image processing DSL --
   DSL compiler
*)

open Imgdsl                             (* DSL definition *)
open Grayimg                            (* Basic image processing lib *)

open Print_code

(* 
#load "bigarray.cma";;
#load "grayimg.cmo";;
*)

(* Types of staged values in our DSL *)
type svalue = 
  | Int  of int code
  | Bool of bool code
  | Img  of img code

(* The type of expressions: the function that receives
   a svalue associated with `it' and computes a svalue
*)
type exp = svalue -> svalue

(* ZZZ Notice how we do the type checking! Type checking just falls out *)

let int x : exp = fun _ -> Int .<x>.

let load x : exp = fun _ -> Img (genlet .<read_pgm x>.)

let display e : exp = fun itv ->
  match e itv with
  | Img x -> Int (genlet .<Grayimg.display ~cmd: "xv" .~x; 0>.)
  | _     -> failwith "Display: type error"

let it : exp    = fun itv -> itv

let binintop op x y : exp = fun itv ->
  match (x itv, y itv) with
  | (Int x, Int y) -> op x y
  | _              -> failwith "binop: type error"

let ( +% ) = binintop (fun x y -> Int .<.~x + .~y>.)
let ( -% ) = binintop (fun x y -> Int .<.~x - .~y>.)
let ( *% ) = binintop (fun x y -> Int .<.~x * .~y>.)
let ( <% ) = binintop (fun x y -> Bool .<.~x < .~y>.)
let ( >% ) = binintop (fun x y -> Bool .<.~x > .~y>.)
let ( =% ) = binintop (fun x y -> Bool .<.~x = .~y>.)

let if_ e et ef : exp = fun itv ->
  match (e itv,et itv,ef itv) with
  | (Bool x,Int et, Int ef) ->
      Int .<if .~x then .~et else .~ef>.
                                (* ZZZ Need more cases *)
  | _          -> failwith "if: type error"

let iterate e body : exp = fun itv ->
  match e itv with
  | Img img -> Img (genlet @@
      .<let pxs = pixels .~img in
        let open Bigarray.Array1 in
        for i = 0 to dim pxs - 1 do
          let p = unsafe_get pxs i in
          .~(match body (Int .<p>.) with
             | Int v -> .<unsafe_set pxs i .~v>.
             | _     -> failwith "iterate: updater must return an int")
        done;
        .~img>.)
  | _     -> failwith "iterate: first argument must be an image"



(* The whole program must produce an int (we don't have void) *)
let run : exp -> int code = fun e -> 
  match e (Int .<0>.) with
  | Int x -> x
  | _     -> failwith "The whole program must produce an int"


(* Try the native compilation *)

(* Save the generated code expression into a file. 
*)
let write_code : string -> 'a code -> unit =
  fun file_name c ->
    let cde = close_code c in           (* make sure the code is closed *)
    let cout = open_out file_name in
    let ppf = Format.formatter_of_out_channel cout in
    (* print code *)
    let () = Print_code.format_code ppf cde in
    let () = Format.fprintf ppf "%!" in
    close_out cout

let run_natively : ?compiler:string -> 'a code -> unit = 
  fun ?(compiler:string="ocamlopt") c ->
  let fname = Filename.temp_file "gen" ".ml" in
  write_code fname c;
  let retc = Sys.command (compiler ^ " bigarray.cmxa grayimg.cmx " ^ fname) in
  let _ = Printf.printf "compilation finished with the code %d\n%!" retc in
  if retc = 0 then begin
    Printf.printf "Press enter to run\n%!";
    ignore (read_line ());
    ignore (Sys.command "./a.out");
    Sys.remove fname
  end


(* A better way: use tagless-final approach.
   It would remove a few assumptions, like that a program must 
   produce an int.
*)

